# -*- coding: utf-8 -*-
#
#  MemoEditor.rb
#  MemoStack
#
#  Created by Fujimoto Hisa on 09/11/06.
#  Copyright (c) 2009 Fujimoto Hisa. All rights reserved.
#

class MemoEditor
  attr_writer   :textView
  attr_accessor :window
  attr_reader   :memo
  
  def self.editors
    if not defined? @@editors
    then  @@editors = []
    else  @@editors
    end
  end
  
  def self.find_editor_for memo
    editors.find{|e| e.memo.memo_id == memo.memo_id }
  end
  
  def self.open(memo, opts={})
    editor = find_editor_for(memo)
    if editor.nil? then
      editor = self.alloc.initWithMemo(memo, options:opts)
      editors << editor
    end
    editor.window.makeKeyAndOrderFront(self)
  end
  
  def initWithMemo(memo, options:opts)
    @memo = memo
    @editable = if opts.has_key? :editable
                then opts[:editable]
                else true
                end
    NSBundle.loadNibNamed("MemoEditorWindow", owner:self)
    self
  end
  
  def awakeFromNib
    @textView.setEditable(@editable)
    load_memo
  end
  
  def toggleEditable(sender)
    @editable = ! @editable
    @textView.setEditable(@editable)
  end
  
  # as NSText delegate
  def textDidChange(ntf)
    @dirty = true
  end
  
  # as NSWindow delegate
  def windowWillClose(ntf)
    save_memo if need_update?
    self.class.editors.delete(self)
  end
  
  private
  
  def need_update?
    ! empty? and @dirty
  end
  
  def empty?
    (@textView.string =~ /\A\s*\z/) ? true : false
  end
  
  def load_memo
    case @memo.kind
    when :webarchive then load_webarchive
    when :txt        then load_text
    else                  load_text
    end
    @window.title = @memo.title
    @dirty = false
  end
  
  def save_memo
    @memo.text = rawdata_of_text
    @memo.data = rawdata_of_webarchive
    @memo.kind = :webarchive
    @memo.update!
    @window.title = @memo.title
    @dirty = false
  end

  ####

  def load_text
    if text = @memo.text then
      @textView.string = text
    end
  end

  def load_webarchive
    if rawdata = @memo.data then
      data = NSData.dataWithBytes(rawdata, length:rawdata.size)
      opts = {}
      dict = Pointer.new_with_type('@') # NSDictionary
      err  = Pointer.new_with_type('@') # NSError
      astr = NSAttributedString.alloc.
        initWithData(data, options:opts, documentAttributes:dict, error:err)
      if astr.nil? then
        emsg = err[0].localizedDescription
        dmesg_fl __FILE__, __LINE__, emsg
        raise "read_webarchive error -- #{emsg}"
      end
      @textView.textStorage.setAttributedString(astr)
    end
  end

  def rawdata_of_webarchive
    astr  = @textView.textStorage
    range = NSRange.new(0, astr.length)
    dict  = { NSDocumentTypeDocumentAttribute=>NSWebArchiveTextDocumentType }
    err   = Pointer.new_with_type('@')
    data  = astr.dataFromRange(range, documentAttributes:dict, error:err)
    if data.nil? then
      msg = err[0].localizedDescription
      raise "rawdata_of_webarchive error -- #{msg}"
    end
    rawdata = ' ' * data.length # データ長の長さの文字列(バイト列)を準備
    data.getBytes(rawdata, length:rawdata.size)
    rawdata
  end

  def rawdata_of_text
    @textView.string
  end
end
