# -*- coding:utf-8 ; indent-tabs-mode:nil -*-
# memo_stack.rb

require 'fileutils'
require 'time'
require 'date'
require 'thread'

module MemoStack
  def MemoStack.memo_stack(kind, opts={})
    case kind
    when :local then LocalMemoStack.new(opts)
    else
      raise "Unknown kind -- #{kind}"
    end
  end
end

# MemoStack.memo_stack のショートカット
def MemoStack(kind, opts={})
  MemoStack.memo_stack(kind, opts)
end

# メモ (実際の操作はスタックに委譲)
class MemoStack::Memo
  attr_reader :memo_id

  def initialize(stack, opts={})
    @stack   = stack
    @memo_id = opts[:memo_id]
  end

  # 便宜上のタイトルを自動生成する
  def title
    @stack.title_for(self)
  end

  def created_at
    @stack.ctime_for(self)
  end

  def modified_at
    @stack.mtime_for(self)
  end

  def update!(text=nil, params={})
    text ||= @text
    params[:data] ||= @data
    params[:kind] ||= @kind
    @stack.save_for(self, text, params)
  end

  def text
    @stack.text_for(self) || @text
  end

  def text=(val)
    @text = val.to_s
  end

  def data
    @stack.data_for(self) || @data
  end

  def data=(val)
    @data = val
  end

  def kind
    @stack.kind_for(self) || @kind
  end

  def kind=(val)
    @kind = val.to_sym
  end
end

class LocalMemoStack

  # ローカルメモスタックディレクトリの構成
  #  #{base_path}/
  #  #{base_path}/METAINFO  -- スタック自体に関する情報
  #  #{base_path}/MEMOPOOL  -- メモ置き場
  #
  def initialize(opts={})
    @mutex = Mutex.new
    if opts.has_key? :path
    then  path = opts[:path]
    else  raise ArgumentError, "required path"
    end
    if File.exist?(path)
      if FileTest.directory?(path)
      then
        @base_path = path
        @meta_path = File.join(path, 'METAINFO')
        @pool_path = File.join(path, 'MEMOPOOL')
      else
        raise "the specified path is not unsupported MemoStack type"
      end
    else
      @base_path = path
      @meta_path = File.join(path, 'METAINFO')
      @pool_path  = File.join(path, 'MEMOPOOL')
      FileUtils.mkdir_p @meta_path
      FileUtils.mkdir_p @pool_path
    end
  end

  # 新規メモの作成(オンメモリ)
  def create_memo
    ::MemoStack::Memo.new(self)
  end

  #### メモの検索・取得するメソッド -- いろんな検索/問合せパターンで取
  #### 得できるようにしたいところ
  
  # 最後に生成したメモ
  def latest_memo
    mid = latest_mid
    mid && ::MemoStack::Memo.new(self, :memo_id=>mid)
  end

  ##### 以下はメモから呼ばれるメソッド (publicだが直接使わない) ####

  # 便宜上のタイトルを自動生成する
  def title_for(memo)
    if memo.memo_id.nil? then
      Time.now.strftime('<%Y-%m-%d %H:%M:%S>')
    else
      dstr, tstr, serial = parts_of_memo_id(memo.memo_id)
      dstr = "#{dstr[0,4]}-#{dstr[4,2]}-#{dstr[6,2]}"
      tstr = "#{tstr[0,2]}:#{tstr[2,2]}:#{tstr[4,2]}"
      "#{dstr} #{tstr} (#{serial})"
    end
  end

  # メモの作成時間 (メモが存在しない場合はnilを返す)
  def ctime_for(memo)
    path = memo_path_for(memo)
    path && File.stat(path).ctime
  end

  # メモの更新時間 (メモが存在しない場合はnilを返す)
  def mtime_for(memo)
    path = memo_path_for(memo)
    path && File.stat(path).mtime
  end

  # メモの更新 (データベースに新規作成・または更新)
  def save_for(memo, text, params={})
    @mutex.synchronize do
      if memo.memo_id
      then  update_for(memo, text, params) # 更新
      else  create_for(memo, text, params) # 新規作成
      end
    end
  end

  # メモテキスト取得
  def text_for(memo)
    path = memo_text_path_for(memo)
    path && File.read(path)
  end

  # メモデータ取得
  def data_for(memo)
    path = memo_data_path_for(memo)
    path && File.read(path)
  end

  # メモデータの種類(＝suffix)
  def kind_for(memo)
    path = memo_data_path_for(memo)
    path && path.split('.').last.to_sym
  end

  private

  # メモの新規作成
  def create_for(memo, text, params)
    mid = create_unique_memo_id
    dir = File.join(@pool_path, mid)
    data, kind = parse_data_param(params)
    text_path = File.join(dir, "#{mid}.txt")          if text
    data_path = File.join(dir, "data-#{mid}.#{kind}") if data
    FileUtils.mkdir_p(dir)
    begin
      File.open(text_path, "w") { |io| io.write(text) } if text
      File.open(data_path, "w") { |io| io.write(data) } if data
      # メモIDをセット (ここ以外では変更しない)
      memo.instance_eval { @memo_id = mid }
    rescue Exception => err
      FileUtils.rm_rf(dir)
      memo.instance_eval { @memo_id = nil }
      raise
    end
  end

  # メモの内容更新
  def update_for(memo, text, params)
    mid = memo.memo_id
    dir  = File.join(@pool_path, mid)

    # 古いデータファイルを .old に移動 (とりあえず安直なバックアッ
    # プ...gitとかを使いたいところ)
    FileUtils.mkdir_p "#{dir}/.old"
    FileUtils.rm_rf   "#{dir}/.old/*"
    items = Dir[ "#{dir}/*" ]
    FileUtils.cp(items, "#{dir}/.old/", :preserve=>true)

    # 新しいデータを保存
    data, kind = parse_data_param(params)
    text_path = File.join(dir, "#{mid}.txt")          if text
    data_path = File.join(dir, "data-#{mid}.#{kind}") if data
    File.open(text_path, "w") { |io| io.write(text) } if text
    File.open(data_path, "w") { |io| io.write(data) } if data
  end

  def parse_data_param(params)
    data = params[:data]
    kind = params[:kind]
    if data and kind.nil? then
      raise "requrie a kind option for the data"
    end
    if kind and data.nil? then
      raise "requrie a data option for the kind"
    end
    [ data, kind ]
  end

  # メモが存在する場合、ディレクトリへのパスを返す。ない場合はnil
  def memo_path_for(memo)
    mid = memo.memo_id
    mid && File.join(@pool_path, mid)
  end

  # メモが存在する場合、必須テキストへのパスを返す。ない場合はnil
  def memo_text_path_for(memo)
    if dir = memo_path_for(memo)
      path = File.join(dir, "#{memo.memo_id}.txt")
      if FileTest.file? path then
        path
      elsif File.exist? path then
        raise "memo text invalid -- #{memo.memo_id}"
      else
        nil
      end
    else
      nil
    end
  end

  # メモが存在する場合、データへのパスを返す。ない場合はnil
  def memo_data_path_for(memo)
    if dir = memo_path_for(memo)
      candidates = Dir[ "#{dir}/data-#{memo.memo_id}.*" ]
      case candidates.size
      when 0 then nil
      when 1 then candidates.first
      else        raise "memo data invalid -- #{memo.memo_id}"
      end
    else
      nil
    end
  end

  # メモの新規作成時に memo_id を生成する
  def create_unique_memo_id
    dtstr = Time.now.strftime("%Y%m%d-%H%M%S")
    snum  = next_serial_number
    mid = "#{dtstr}-%04d" % snum
    while File.exist? "#{@pool_path}/#{mid}" do
      snum += 1
      mid = "#{dtstr}-%04d" % snum
    end
    mid
  end

  # 最新のn件
  def latest_mids(n)
    Dir.chdir(@pool_path) { Dir["*-*-*"] }.sort.reverse[0,n]
  end
  
  def latest_mid
    Dir.chdir(@pool_path) { Dir["*-*-*"] }.sort.last
  end
  

  # その日に作られたメモすべて
  def todays_memo_ids
    datestr = Date.today.strftime("%Y%m%d")
    Dir.chdir(@pool_path) { Dir["#{datestr}-*-*"] }
  end

  # 次のシリアル番号(毎日ふられる)候補
  def next_serial_number
    if last_mid = todays_memo_ids.sort.last
    then serial_number_of_memo_id(last_mid) + 1
    else 0
    end
  end

  # メモIDを [日時,時間,シリアル番号] に分解する
  def parts_of_memo_id(mid)
    ma = /\A(\d{8})-(\d{6})-(\d+)\Z/.match(mid)
    [ ma[1], ma[2], ma[3] ]
  end

  # 指定されたメモIDの日付成分
  def datestr_of_memo_id(mid) parts_of_memo_id(mid)[0] end

  # 指定されたメモIDの時間成分
  def tiemstr_of_memo_id(mid) parts_of_memo_id(mid)[1] end

  # 指定されたメモIDに含まれるシリアル番号を取得
  def serial_number_of_memo_id(mid)
    parts_of_memo_id(mid)[2].to_i
  end
end


if __FILE__ == $0 then
  require 'test/unit'
  
  class TestMemoStack < Test::Unit::TestCase
    
    def setup
      @stack_path = "/tmp/hogehoge.memostack"
      @stack = MemoStack(:local, :path=>"/tmp/hogehoge.memostack")
    end
    
    def teardown
      remove_local_stack(@stack_path)
    end
    
    # ローカルディスク上にメモスタックを新規生成
    def test_create_local_memo_stack
      stack_path = "/tmp/#{Time.now.to_f.to_s}.memostack"
      
      assert_equal( false, 
                    File.exist?(stack_path),
                    'メモスタックは存在しない')
      
      assert_respond_to( MemoStack(:local, :path=>stack_path),
                         :create_memo,
                         '新規メモスタックを生成')
      
      assert_equal( true, 
                    File.exist?(stack_path),
                    'メモスタックが新しく作られた')
    ensure
      remove_local_stack(stack_path) # 後始末
    end
    
    # ローカルディスク上の既存のメモスタック
    def test_create_local_memo_stack
      stack_path = "/tmp/#{Time.now.to_f.to_s}.memostack"
      MemoStack(:local, :path=>stack_path)
      
      assert_equal( true, 
                    File.exist?(stack_path),
                    'メモスタックが存在')

      assert_respond_to( MemoStack(:local, :path=>stack_path),
                         :create_memo,
                         '既存のメモスタック')
    ensure
      remove_local_stack(stack_path) # 後始末
    end
    
    
    def test_create_memo
      assert_nothing_raised { @stack.create_memo }
    end
    
    def test_memo_update!
      memo = @stack.create_memo

      assert_nothing_raised 'テキスト(必須)のみ保存' do 
        memo.text = "hello, world (text only)"
        memo.update!
      end
      assert_equal("hello, world (text only)", memo.text)
      assert_equal(nil, memo.data)
      assert_equal(nil, memo.kind)

      assert_nothing_raised 'テキスト(必須)のみ保存 (2)' do 
        memo.update!("hello, world (text only 2)")
      end
      assert_equal("hello, world (text only 2)", memo.text)
      assert_equal(nil, memo.data)
      assert_equal(nil, memo.kind)

      assert_nothing_raised '別形式のデータも保存' do
        memo.text = "hello, world (text)"
        memo.data = 'hello, world (data)'
        memo.kind = :txt
        memo.update!
      end
      assert_equal("hello, world (text)", memo.text)
      assert_equal('hello, world (data)', memo.data)
      assert_equal(:txt, memo.kind)

      assert_nothing_raised '別形式のデータも保存 (2)' do
        memo.update!("hello, world (text 2)", 
                     :data=>'hello, world (data 2)', :kind=>:txt)
      end
      assert_equal("hello, world (text 2)", memo.text)
      assert_equal('hello, world (data 2)', memo.data)
      assert_equal(:txt, memo.kind)
    end

    # (便宜上の)タイトルを自動生成する
    def test_memo_title
      memo = @stack.create_memo
      assert_respond_to(memo, :title)
      assert_kind_of(String, memo.title)
      memo.update!(:data=>"hoge", :kind=>:txt)
      assert_kind_of(String, memo.title)
    end

    def test_memo_created_at
      memo = @stack.create_memo
      assert_respond_to(memo, :created_at)
      assert_equal(nil, memo.created_at, 'update!前なのでまだ保存されていない')
      memo.update!(:data=>"hoge", :kind=>:txt)
      assert_kind_of(Time, memo.created_at, 'update!後は作成時間を返す')
    end

    def test_memo_modified_at
      memo = @stack.create_memo
      assert_respond_to(memo, :modified_at)
      assert_equal(nil, memo.modified_at, 'update!前なのでまだ保存されていない')
      memo.update!(:data=>"hoge", :kind=>:txt)
      assert_kind_of(Time, memo.modified_at, 'update!後は更新時間を返す')
    end

    def test_memo_text
      memo = @stack.create_memo
      assert_respond_to(memo, :text)
      assert_equal(nil, memo.text, 'update!前なのでまだ保存されていない')
      memo.update!('hello, world')
      assert_equal("hello, world", memo.text)
    end

    def test_memo_data
      memo = @stack.create_memo
      assert_respond_to(memo, :data)
      assert_equal(nil, memo.data, 'update!前なのでまだ保存されていない')
      memo.update!('hello, world')
      assert_equal(nil, memo.data, '必須テキスト以外のデータはない')
      memo.update!('hello, world', :data=>'aaa', :kind=>:txt)
      assert_equal('aaa', memo.data)
    end

    def test_memo_kind
      memo = @stack.create_memo
      assert_respond_to(memo, :kind)
      assert_equal(nil, memo.kind, 'update!前なのでまだ保存されていない')
      memo.update!('hello, world')
      assert_equal(nil, memo.kind, '必須テキスト以外のデータはない')
      memo.update!('hello, world', :data=>'aaa', :kind=>:txt)
      assert_equal(:txt, memo.kind)
    end

    def test_latest_memo
      assert_respond_to(@stack, :latest_memo)
      memo = @stack.create_memo
      memo.update!('hello, world', :data=>'aaa', :kind=>:txt)
      assert_nothing_raised { memo = @stack.latest_memo }
      assert_equal('hello, world', memo.text)
      assert_equal('aaa',          memo.data)
      assert_equal(:txt,           memo.kind)
    end

    private
    def remove_local_stack(path)
      FileUtils.rm_rf(path)
    end
  end
end
